from __future__ import annotations

import importlib
import importlib.util
from dataclasses import dataclass, field
from pathlib import Path
from types import ModuleType
from typing import Optional, Type

from mcanitexgen import utils

from .parser import Action, Duration, Sequence, SequenceAction, State, Timeframe


class GeneratorError(Exception):
    pass


def load_animations_from_file(path: Path):
    spec = importlib.util.spec_from_file_location(path.name, path)
    if not spec:
        raise GeneratorError(f"Couldn't load animations from '{path}'")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return get_texture_animations_from_module(module)


def get_texture_animations_from_module(module: ModuleType):
    return {
        k: v
        for k, v in module.__dict__.items()
        if isinstance(v, Type) and issubclass(v, TextureAnimation) and v != TextureAnimation
    }


def animation(texture: Path, main_sequence: str = "main", interpolate=False, frametime=1):
    def wrapper(cls: Type[TextureAnimation]):
        cls.sequences = {}
        cls.states = {}
        for name, val in cls.__dict__.items():
            if isinstance(val, State):
                cls.states[val.index] = val
            elif isinstance(val, Sequence):
                val.name = name
                cls.sequences[name] = val
        cls.root = cls.sequences[main_sequence]

        cls.texture = texture if isinstance(texture, Path) else Path(texture)
        cls.interpolate = interpolate
        cls.frametime = frametime

        animation = unweighted_sequence_to_animation(cls.root, 0)
        cls.start = animation.start
        cls.end = animation.end
        cls.frames = animation.frames
        cls.marks = animation.marks
        return cls

    return wrapper


class TextureAnimation:
    texture: Path
    interpolate: bool
    frametime: int

    sequences: dict[str, Sequence]
    states: dict[int, State]
    root: Sequence

    start: int
    end: int
    frames: list[dict]
    marks: dict[str, Mark]


@dataclass
class Animation:
    start: int
    end: int
    frames: list[dict] = field(default_factory=list)
    marks: dict[str, Mark] = field(default_factory=dict)

    def append(self, other: Animation):
        # Fill time gap between animations
        time_gap = other.start - self.end
        if time_gap > 0 and self.frames:
            self.frames[-1]["time"] += time_gap
        elif time_gap < 0:
            raise GeneratorError(
                f"Can't append to animation that starts before the other ends"
            )

        self.end = other.end
        self.frames += other.frames
        self.marks |= other.marks

    def add_frame(self, index: int, start: int, end: int):
        if end - start <= 0:
            raise GeneratorError(f"Illegal start and end for frame: '{start}' '{end}'")

        if len(self.frames) == 0:
            # Animation starts at start of first frame
            self.start = start
        elif start - self.end > 0:
            # Extend time of the last frame to fill the gap to the new frame
            self.frames[-1]["time"] += start - self.end

        self.end = end
        self.frames.append({"index": index, "time": end - start})


@dataclass
class Mark:
    start: int
    end: int


def unweighted_sequence_to_animation(sequence: Sequence, start: int):
    assert not sequence.is_weighted
    animation = Animation(start, start)

    for action in sequence.actions:
        action_start, action_duration = get_unweighted_action_timeframe(action, animation.end)
        append_action_to_animation(action, action_start, action_duration, animation)

    return animation


def weighted_sequence_to_animation(sequence: Sequence, start: int, duration: int):
    assert sequence.is_weighted
    animation = Animation(start, start)

    distributable_duration = duration - sequence.constant_duration

    if distributable_duration <= 0:
        raise GeneratorError(
            f"Duration '{duration}' is not enough for the weighted sequence '{sequence.name}' with constant duration '{sequence.constant_duration}'"
        )

    duration_distributor = utils.DurationDistributor(
        distributable_duration, sequence.total_weight
    )

    for action in sequence.actions:
        if action.is_weighted:
            action_start = animation.end
            action_duration = duration_distributor.take(int(action.time))
        else:
            action_start, action_duration = get_unweighted_action_timeframe(
                action, animation.end
            )

        append_action_to_animation(action, action_start, action_duration, animation)

    if not duration_distributor.is_empty():
        raise GeneratorError(f"Couldn't distribute duration over weights")

    return animation


def get_unweighted_action_timeframe(action: Action, action_start: int):
    if isinstance(action.time, Duration):
        action_duration = int(action.time)
    elif isinstance(action.time, Timeframe):
        start, end, duration = action.time.start, action.time.end, action.time.duration
        if start and end and duration:
            action_start = start
            action_duration = duration
        elif not start and end and not duration:
            action_duration = end - action_start
        else:
            raise GeneratorError(
                f"Unexpected combination of start, end, duration: '{start}', '{end}', '{duration}'"
            )
    else:
        action_duration = None

    return (action_start, action_duration)


def append_action_to_animation(
    action: Action, start: int, duration: Optional[int], anim: Animation
):
    if isinstance(action, SequenceAction):
        anim.append(sequence_action_to_animation(action, start, duration))
    else:
        assert duration is not None
        anim.add_frame(action.state.index, start, start + duration)

    # Add mark
    if action.mark:
        anim.marks[action.mark] = Mark(start, anim.end)


def sequence_action_to_animation(action: SequenceAction, start: int, duration: Optional[int]):
    anim = Animation(start, start)

    if action.sequence.is_weighted:
        if not duration:
            raise GeneratorError(
                f"Didn't pass duration to weighted sequence '{action.sequence.name}'"
            )

        if action.is_weighted:
            duration_distributor = utils.DurationDistributor(duration, action.repeat)
            for _ in range(action.repeat):
                anim.append(
                    weighted_sequence_to_animation(
                        action.sequence, anim.end, duration_distributor.take(1)
                    )
                )

            if not duration_distributor.is_empty():
                raise GeneratorError(f"Couldn't distribute duration over weights")
        else:
            for _ in range(action.repeat):
                anim.append(
                    weighted_sequence_to_animation(action.sequence, anim.end, duration)
                )
    else:
        if duration:
            raise GeneratorError(
                f"Passing duration to unweighted sequence '{action.sequence.name}'"
            )

        for _ in range(action.repeat):
            anim.append(unweighted_sequence_to_animation(action.sequence, anim.end))

    return anim



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Optional, Union


class ParserError(Exception):
    pass


@dataclass(init=False)
class State:
    index: int

    def __init__(self, index):
        self.index = index

    def __call__(
        self,
        start: Optional[int] = None,
        end: Optional[int] = None,
        duration: Optional[int] = None,
        weight: Optional[int] = None,
        mark: Optional[str] = None,
    ):
        time = Time.from_args(start, end, duration, weight)
        if time is None:
            time = Duration(1)

        return StateAction(self, time, mark)


@dataclass(init=False)
class Sequence:
    actions: list[Action]
    name: Optional[str] = None

    def __init__(self, *actions: Union[Action, Sequence]):
        self.actions = list(
            map(
                lambda a: a() if (isinstance(a, Sequence) or isinstance(a, State)) else a,
                actions,
            )
        )

        self.total_weight = sum(map(lambda a: int(a.time), self.weighted_actions()))
        self.is_weighted = self.total_weight > 0

        self.constant_duration = sum(map(lambda a: a.constant_duration(), self.actions))

    def weighted_actions(self):
        return filter(lambda a: a.is_weighted, self.actions)

    def __call__(
        self,
        repeat=1,
        start: Optional[int] = None,
        end: Optional[int] = None,
        duration: Optional[int] = None,
        weight: Optional[int] = None,
        mark: Optional[str] = None,
    ):
        time = Time.from_args(start, end, duration, weight)
        return SequenceAction(self, time, repeat, mark)

    def __mul__(self, other):
        if isinstance(other, int):
            return self(repeat=other)
        else:
            raise NotImplementedError()

    def __rmul__(self, other):
        return self.__mul__(other)


class Time:
    def from_args(
        start: Optional[int] = None,
        end: Optional[int] = None,
        duration: Optional[int] = None,
        weight: Optional[int] = None,
    ):
        if weight != None:
            if weight <= 0:
                raise ParserError(f"Weight of time must be at least 1")
            if start or end or duration:
                raise ParserError(f"Weighted time can't have start, end or duration")
            return Weight(weight)
        elif start != None or end != None:
            return Timeframe(start, end, duration)
        elif duration != None:
            if duration <= 0:
                raise ParserError(f"Duration must be at least 1")
            return Duration(duration)
        else:
            return None


class Duration(int, Time):
    pass


class Weight(int, Time):
    pass


@dataclass
class Timeframe(Time):
    start: Optional[int] = None
    end: Optional[int] = None
    duration: Optional[int] = None

    def __post_init__(self):
        # Check for illegal inputs
        if self.start is None and self.end is None and self.duration is None:
            raise ParserError(f"Timeframe must have at least one of start, end, duration set")
        elif self.start is None and self.end != None and self.duration != None:
            raise ParserError(f"Timeframes without start can't have end and duration")

        if self.start != None:
            # Make sure start, end and duration are set
            if self.end is None or self.duration is None:
                self._deduce_unset_attributes()

            if self.end - self.start != self.duration:
                raise ParserError(
                    f"Start, end and duration of timeframe don't match: {self.start}, {self.end}, {self.duration}"
                )

    def _deduce_unset_attributes(self):
        if self.start != None:
            if self.end != None:
                self.duration = self.end - self.start
            elif self.duration != None:
                self.end = self.start + self.duration
            else:
                self.end = self.start + 1
                self.duration = 1


class Action(abc.ABC):
    def __init__(self, time: Optional[Time], mark: Optional[str] = None):
        self.time = time
        self.mark = mark

    @property
    def is_weighted(self):
        return isinstance(self.time, Weight)

    def constant_duration(self):
        if isinstance(self.time, Duration):
            return self.time
        elif isinstance(self, SequenceAction) and not self.is_weighted:
            return self.sequence.constant_duration
        else:
            return 0


@dataclass(init=False)
class StateAction(Action):
    state: State
    time: Time
    mark: Optional[str] = None

    def __init__(self, state: State, time: Time, mark: Optional[str] = None):
        super().__init__(time, mark)
        self.state = state


@dataclass(init=False)
class SequenceAction(Action):
    sequence: Sequence
    time: Optional[Time]
    repeat: int
    mark: Optional[str]

    def __init__(
        self,
        sequence: Sequence,
        time: Optional[Time] = None,
        repeat=1,
        mark: Optional[str] = None,
    ):
        super().__init__(time, mark)
        self.sequence = sequence
        self.repeat = repeat

        if self.repeat <= 0:
            raise ParserError(f"Sequence cannot be repeated '{self.repeat}' times")

    def __mul__(self, other):
        if isinstance(other, int):
            if other <= 0:
                raise ParserError(f"Sequence cannot be repeated '{other}' times")

            self.repeat = other
        else:
            raise NotImplementedError()
        return self

    def __rmul__(self, other):
        return self.__mul__(other)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mcanitexgen']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.1.2,<9.0.0',
 'numpy>=1.20.1,<2.0.0',
 'ruamel.yaml>=0.16.12,<0.17.0',
 'typer>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'mcanitexgen',
    'version': '0.1.0',
    'description': 'A texture animation generator for Minecraft .mcmeta files',
    'long_description': None,
    'author': 'Oran9eUtan',
    'author_email': 'oran9eutan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import math
from pathlib import Path
from typing import Generator, Iterable

import ruamel.yaml as yaml


def load_yaml_file(path: Path):
    with path.open("r", encoding="utf-8") as f:
        return yaml.load(f, Loader=yaml.Loader)


def round_half_away_from_zero(num) -> int:
    """ y = sign(x) * floor(|x|+0.5) """

    if num >= 0:
        return math.floor(num + 0.5)
    else:
        return math.floor(num - 0.5)


def partition_by_weights(
    number: int, total_weight: int, weights: Iterable[int]
) -> Generator[int, None, None]:
    """ Partitions a number into a series of integers proportianal to a series of weights """

    remaining = number
    remaining_weight = total_weight

    for weight in weights:
        weighted_amount = int(
            round_half_away_from_zero((weight * remaining) / remaining_weight)
        )
        remaining -= weighted_amount
        remaining_weight -= weight

        if remaining_weight < 0:
            raise ValueError(f"Weights exceed passed total weight of '{total_weight}'")

        yield weighted_amount


class DurationDistributor:
    def __init__(self, num: int, total_weight: int):
        self.num = num
        self.total_weight = total_weight
        self.remaining = num
        self.remaining_weight = total_weight

    def take(self, weight: int):
        weighted_amount = int(
            round_half_away_from_zero((weight * self.remaining) / self.remaining_weight)
        )
        self.remaining -= weighted_amount
        self.remaining_weight -= weight

        if self.remaining_weight < 0:
            raise ValueError(f"Weights exceed passed total weight of '{self.total_weight}'")

        return weighted_amount

    def is_empty(self):
        return self.remaining <= 0
